# do not import all endpoints into this module because that uses a lot of memory and stack frames
# if you need the ability to import all endpoints from this module, import them with
# from sedaro_base_client.paths.simulations_jobs_job_id_externals_agent_id_external_state_block_id import Api

from sedaro_base_client.paths import PathValues

path = PathValues.SIMULATIONS_JOBS_JOB_ID_EXTERNALS_AGENT_ID_EXTERNAL_STATE_BLOCK_ID