# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: cosim.proto
# Protobuf Python Version: 5.27.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    27,
    2,
    '',
    'cosim.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0b\x63osim.proto\x12\x05\x63osim\"\xec\x01\n\x10SimulationAction\x12\x1e\n\x16\x63luster_handle_address\x18\x01 \x01(\t\x12\x0e\n\x06job_id\x18\x02 \x01(\t\x12\x10\n\x08\x61gent_id\x18\x03 \x01(\t\x12\x1f\n\x17\x65xternal_state_block_id\x18\x04 \x01(\t\x12\x16\n\ttimestamp\x18\x05 \x01(\x01H\x01\x88\x01\x01\x12!\n\x07\x63onsume\x18\x06 \x01(\x0b\x32\x0e.cosim.ConsumeH\x00\x12!\n\x07produce\x18\x07 \x01(\x0b\x32\x0e.cosim.ProduceH\x00\x42\t\n\x07requestB\x0c\n\n_timestamp\"\xe6\x01\n\x12SimulationResponse\x12\x1b\n\x05state\x18\x01 \x01(\x0e\x32\x0c.cosim.State\x12\x11\n\ttimestamp\x18\x02 \x01(\x01\x12\x32\n\x10\x63onsume_response\x18\x03 \x01(\x0b\x32\x16.cosim.ConsumeResponseH\x00\x12\x32\n\x10produce_response\x18\x04 \x01(\x0b\x32\x16.cosim.ProduceResponseH\x00\x12,\n\rfail_response\x18\x05 \x01(\x0b\x32\x13.cosim.FailResponseH\x00\x42\n\n\x08response\"\x18\n\x07\x43onsume\x12\r\n\x05index\x18\x01 \x01(\x05\"/\n\x0f\x43onsumeResponse\x12\r\n\x05index\x18\x01 \x01(\x01\x12\r\n\x05value\x18\x02 \x01(\t\"\'\n\x07Produce\x12\r\n\x05index\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\t\" \n\x0fProduceResponse\x12\r\n\x05index\x18\x01 \x01(\x05\"\x1f\n\x0c\x46\x61ilResponse\x12\x0f\n\x07message\x18\x01 \x01(\t\"?\n\tAuthorize\x12\x12\n\nauth_token\x18\x01 \x01(\t\x12\x1e\n\x16\x63luster_handle_address\x18\x02 \x01(\t\":\n\x11\x41uthorizeResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x14\n\x0csession_uuid\x18\x02 \x01(\t*!\n\x05State\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\x32\x8a\x01\n\x05\x43osim\x12\x44\n\x0eSimulationCall\x12\x17.cosim.SimulationAction\x1a\x19.cosim.SimulationResponse\x12;\n\rAuthorizeCall\x12\x10.cosim.Authorize\x1a\x18.cosim.AuthorizeResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'cosim_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_STATE']._serialized_start=802
  _globals['_STATE']._serialized_end=835
  _globals['_SIMULATIONACTION']._serialized_start=23
  _globals['_SIMULATIONACTION']._serialized_end=259
  _globals['_SIMULATIONRESPONSE']._serialized_start=262
  _globals['_SIMULATIONRESPONSE']._serialized_end=492
  _globals['_CONSUME']._serialized_start=494
  _globals['_CONSUME']._serialized_end=518
  _globals['_CONSUMERESPONSE']._serialized_start=520
  _globals['_CONSUMERESPONSE']._serialized_end=567
  _globals['_PRODUCE']._serialized_start=569
  _globals['_PRODUCE']._serialized_end=608
  _globals['_PRODUCERESPONSE']._serialized_start=610
  _globals['_PRODUCERESPONSE']._serialized_end=642
  _globals['_FAILRESPONSE']._serialized_start=644
  _globals['_FAILRESPONSE']._serialized_end=675
  _globals['_AUTHORIZE']._serialized_start=677
  _globals['_AUTHORIZE']._serialized_end=740
  _globals['_AUTHORIZERESPONSE']._serialized_start=742
  _globals['_AUTHORIZERESPONSE']._serialized_end=800
  _globals['_COSIM']._serialized_start=838
  _globals['_COSIM']._serialized_end=976
# @@protoc_insertion_point(module_scope)
